#!/usr/bin/perl

use CGI::Carp qw(fatalsToBrowser);

use strict;

use CGI;
use DBI;


BEGIN { require "config.pl"; }
BEGIN { require "templates.pl"; }
BEGIN { require "wakautils.pl"; }



my $dbh=DBI->connect(SQL_DBI_SOURCE,SQL_USERNAME,SQL_PASSWORD,{AutoCommit=>1}) or die "Couldn't connect to SQL database";

check_database();

my $query=new CGI;
my $task=$query->param("task");

if(!$task)
{
	my $image=$query->param("image");

	if($image) { show_image($image) }
	else { show_frontpage() }
}
elsif($task eq "history")
{
	my $image=$query->param("image");
	my $id=$query->param("id");
	my $timestamp=$query->param("timestamp");

	show_history($image,$id,$timestamp);
}
elsif($task eq "newnotes")
{
	my $image=$query->param("image");
	my $mintime=$query->param("mintime");

	return_new_notes($image,$mintime);
}
elsif($task eq "save")
{
	my $image=$query->param("image");
	my $id=$query->param("id");
	my $text=$query->param("text");
	my $x=$query->param("x");
	my $y=$query->param("y");
	my $w=$query->param("w");
	my $h=$query->param("h");
	my $color1=$query->param("color1");
	my $color2=$query->param("color2");
	my $name=$query->param("name");
	my $oldtime=$query->param("oldtime");
	my $mintime=$query->param("mintime");

	my $message=save_note($image,$id,$text,$x,$y,$w,$h,$color1,$color2,$name,$oldtime);

	return_new_notes($image,$mintime,$id,$message?(message=>js_string($message)):());
}
elsif($task eq "delete")
{
	my $image=$query->param("image");
	my $id=$query->param("id");
	my $name=$query->param("name");
	my $oldtime=$query->param("oldtime");
	my $mintime=$query->param("mintime");

	my $message=save_note($image,$id,"### DELETED ###",0,0,0,0,"deleted","deleted",$name,$oldtime);

	return_new_notes($image,$mintime,undef,$message?(message=>js_string($message)):());
}
elsif($task eq "revert")
{
	my $image=$query->param("image");
	my $id=$query->param("id");
	my $timestamp=$query->param("timestamp");
	my $name=$query->param("name");

	revert_note($image,$id,$timestamp,$name);

	print "Status: 302 Moved\n";
	print "Location: $ENV{SCRIPT_NAME}?image=".urlenc($image)."\n";
	print "\n";
}



#
# 
#

sub show_frontpage()
{
	my ($sth,$row,@images);

	$dbh->do("CREATE TEMPORARY TABLE temp AS ".
	"SELECT url AS maxurl,MAX(timestamp) AS maxtime FROM entries GROUP BY url;") or die $dbh->errstr();

	$sth=$dbh->prepare("SELECT * FROM entries INNER JOIN temp ".
	"ON url=maxurl AND timestamp=maxtime ORDER BY maxtime DESC;") or die $dbh->errstr();
	$sth->execute() or die $sth->errstr();

	while($row=$sth->fetchrow_hashref()) { push @images,$row }

	make_html_header();

	print FRONTPAGE_TEMPLATE->(
		images=>\@images,
	);
}

sub show_image($)
{
	my ($image)=@_;

	$image=clean_string($image);

	my @notes=get_notes($image,0);

	make_html_header();

	print IMAGE_TEMPLATE->(
		image=>$image,
		notes=>\@notes,
		jsnotes=>make_js(@notes),
	);
}

sub show_history($$$)
{
	my ($image,$id,$timestamp)=@_;
	my ($sth,$row,@edits);

	$image=clean_string($image);
	$id=clean_string($id);
	$timestamp=clean_string($timestamp);

	if($timestamp)
	{
		$sth=$dbh->prepare("SELECT * FROM entries WHERE url=? AND id=? AND timestamp=?;") or die $dbh->errstr();
		$sth->execute($image,$id,$timestamp) or die $sth->errstr();
	}
	else
	{
		$sth=$dbh->prepare("SELECT * FROM entries WHERE url=? AND id=? ORDER BY timestamp DESC;") or die $dbh->errstr();
		$sth->execute($image,$id) or die $sth->errstr();
	}

	while($row=$sth->fetchrow_hashref()) { push @edits,$row }

	make_html_header();

	print HISTORY_TEMPLATE->(
		image=>$image,
		id=>$id,
		edits=>\@edits,
	);
}

sub make_html_header()
{
	print "Content-Type: ".get_xhtml_content_type("utf-8",1)."\n";
	print "\n";
}

sub return_new_notes($$$%)
{
	my ($image,$mintime,$marked,%more)=@_;
	my @notes=get_notes($image,$mintime);

	for(@notes) { $$_{marked}=1 if($$_{id} eq $marked) }

	print "Content-Type: text/javascript\n\n";
	print "(".js_hash(newnotes=>make_js(@notes),%more).")";
}

sub get_notes($$$)
{
	my ($image,$mintime)=@_;
	my ($sth,$row,@notes);

	$dbh->do("CREATE TEMPORARY TABLE temp AS ".
	"SELECT id AS maxid,MAX(timestamp) AS maxtime FROM entries WHERE url=? AND timestamp>? GROUP BY id;",undef,$image,$mintime) or die $dbh->errstr();

	$sth=$dbh->prepare("SELECT * FROM entries INNER JOIN temp ".
	"ON id=maxid AND timestamp=maxtime WHERE entries.url=? ORDER BY timestamp DESC;") or die $dbh->errstr();
	$sth->execute($image) or die $sth->errstr();

	while($row=$sth->fetchrow_hashref()) { push @notes,$row }

	return @notes;
}

use Encode qw(decode);

sub make_js(@)
{
	return js_array(map
	{
		my %note=%{$_};

		js_hash(
			id=>js_string($note{id}),
			timestamp=>$note{timestamp},
			text=>clean_to_js(decode("utf8",$note{text})),
			color1=>js_string($note{color1}),
			color2=>js_string($note{color2}),
			x=>$note{x},y=>$note{y},
			w=>$note{w},h=>$note{h},
			name=>clean_to_js(decode("utf8",$note{name})),
			trip=>js_string($note{trip}),
			ip=>js_string(mask_ip($note{ip},SECRET)),
			marked=>js_string($note{marked}),
		);
	} @_);
}


#
# Saving
#

sub save_note($$$$$$$$$$$)
{
	my ($image,$id,$text,$x,$y,$w,$h,$color1,$color2,$name,$oldtime)=@_;
	my ($trip,$sth);

	die if($x=~/[^0-9]/);
	die if($y=~/[^0-9]/);
	die if($w=~/[^0-9]/);
	die if($h=~/[^0-9]/);
	die if($color1=~/[^a-z0-9]/);
	die if($color2=~/[^a-z0-9]/);

	# fix newlines in as many places as possible, goddamn you internet
	$text=~s/\r\n|\r|\n/\n/g;

	$image=clean_string($image);
	$id=clean_string($id);
	$text=clean_string($text);

	($name,$trip)=process_tripcode($name,"!",SECRET,"utf-8");

	my $time=time().sprintf("%03d",rand(1000));
	my $ip=$ENV{REMOTE_ADDR};

	$sth=$dbh->prepare("SELECT MAX(timestamp) FROM entries WHERE url=? AND id=?") or die $dbh->errstr();
	$sth->execute($image,$id) or die $sth->errstr();
	my $newest=($sth->fetchrow_array())[0];

	$dbh->do("INSERT INTO entries VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?);",undef,
	$image,$id,$time,$text,$x,$y,$w,$h,$color1,$color2,$name,$trip,$ip) or die $dbh->errstr();

	return "Somebody else has edited this note while you worked on it.\n\nPlease check the edit history to see what they changed." if($newest and $newest ne $oldtime);
	return "";
}

sub revert_note($$$$)
{
	my ($image,$id,$timestamp,$name)=@_;
	my ($sth,$row);

	$image=clean_string($image);
	$id=clean_string($id);

	$sth=$dbh->prepare("SELECT * FROM entries WHERE url=? AND id=? AND timestamp=?;") or die $dbh->errstr();
	$sth->execute($image,$id,$timestamp) or die $sth->errstr();

	if($row=$sth->fetchrow_hashref())
	{
		save_note($image,$id,$$row{text},$$row{x},$$row{y},$$row{w},$$row{h},$$row{color1},$$row{color2},$name,0);
	}
}



#
# Database utils
#

sub check_database()
{
	if(!table_exists("entries"))
	{
		init_database();
		return;
	}

	if(!column_exists("entries","color1"))
	{
#	die "upgrading!";

		$dbh->do("CREATE TEMPORARY TABLE tmp AS SELECT url,id,timestamp,text,x,y,w,h,? AS color1,? AS color2,name,trip,ip FROM entries;",undef,"yellow","cyan") or die $dbh->errstr();
		init_database();
		$dbh->do("INSERT INTO entries SELECT * FROM tmp;") or die $dbh->errstr();
		return;
	}
}

sub init_database()
{
	$dbh->do("DROP TABLE entries;") if(table_exists("entries"));
	$dbh->do("CREATE TABLE entries (".

	"url TEXT,".
	"id TEXT,".
	"timestamp TEXT,".

	"text TEXT,".
	"x INTEGER,".
	"y INTEGER,".
	"w INTEGER,".
	"h INTEGER,".
	"color1 TEXT,".
	"color2 TEXT,".

	"name TEXT,".
	"trip TEXT,".
	"ip TEXT".

	");") or die $dbh->errstr();
}

sub table_exists($)
{
	my ($table)=@_;
	my ($sth);

	return 0 unless($sth=$dbh->prepare("SELECT * FROM $table LIMIT 1;"));
	return 0 unless($sth->execute());
	return 1;
}

sub column_exists($$)
{
	my ($table,$column)=@_;
	my ($sth);

	return 0 unless($sth=$dbh->prepare("SELECT $column FROM $table LIMIT 1;"));
	return 0 unless($sth->execute());
	return 1;
}
